indexing

	description: "Base class for a single service."

	author: "Berend de Boer"
	copyright:   "Copyright (c) 2003, Berend de Boer"
	license:     "Eiffel Forum Freeware License v2 (see forum.txt)"


deferred class

	NT_SERVICE


inherit

	ANY

	WEL_ERROR
		export
			{NONE} all
		end

	NT_SERVICE_DEBUGGING
		export
			{NONE} all
		end

	WINSVC_CONSTANTS

	WINSVC_FUNCTIONS_EXTERNAL
		export
			{NONE} all
		end

	EWG_EXTERNAL_STRING_ROUTINES
		export
			{NONE} all
		end

	EXCEPTIONS
		export
			{NONE} all
		end


feature {NONE} -- Initialization

	make (an_application: NT_SERVICE_APPLICATION) is
		require
			application_not_void: an_application /= Void
		do
			application := an_application
			create status.make_new_shared
			status.set_dwservicetype (service_win32_own_process)
			status.set_dwcontrolsaccepted (service_accept_stop)
			status.set_dwwaithint (default_wait_hint)
		ensure
			application_set: application = an_application
		end


feature -- Access

	application: NT_SERVICE_APPLICATION
			-- Class that started this service.

	callback_context: POINTER
			-- Context to be passed to `RegisterServiceCtrlHandlerEx'.

	cm_service_control_code: INTEGER is 0x0401
			-- Message code for communication with service thread.
			-- WM_USER + 1

	current_state: INTEGER is
			-- Current state of service.
		do
			Result := status.get_dwcurrentstate
		end

	name: STRING is
		deferred
		ensure
			name_not_empty: Result /= Void and then not Result.is_empty
		end

	status: SERVICE_STATUS_STRUCT
			-- Current status of service.

	status_handle: POINTER
			-- Service status handle if not zero.

	default_wait_hint: INTEGER is
		do
			Result := 5000
		end


feature -- Status

	is_status_handle_valid: BOOLEAN is
			-- Does `status_handle' refer to a valid status handle?
		do
			Result := status_handle /= default_pointer
		end


feature -- Status change

	report_status is
			-- Report status to SCM.
		require
			have_status_handle: is_status_handle_valid
		local
			success: BOOLEAN
		do
			success := set_service_status_external (status_handle, status.item) /= 0
			if not success then
				log ("report_status failed.")
			end
		end


	set_status (new_status: INTEGER) is
			-- Change current status in `status' and report it.
		require
			have_status_handle: is_status_handle_valid
		local
			last_status: INTEGER
		do
			last_status := status.get_dwcurrentstate
			status.set_dwcurrentstate (new_status)
			-- Reset dwCheckPoint and dwWaitHint when we're running
			inspect new_status
			when service_start_pending then
				-- Assume wait hint in struct is set correctly
				if status.get_dwcheckpoint = 0 then
					status.set_dwcheckpoint (default_wait_hint)
				end
				if
					last_status = service_start_pending or else
					last_status = service_stop_pending
				then
					status.set_dwcheckpoint (status.get_dwcheckpoint + 1)
				else
					status.set_dwcheckpoint (0)
				end
			when service_running, service_stopped, service_paused then
				status.set_dwcheckpoint (0)
				status.set_dwwaithint (0)
			else
				-- nothing special
			end
			report_status
		end


feature -- Main

	main (arguments: ARRAY [STRING]) is
			-- Start of service. Called in separate thread.
			-- Should not call any win32 APIs until `current_state' is set to
			-- `service_running', according the the MSDN
			-- documentation. That won't work, I'm afraid.
		do
			debug ("service")
				log ("NT_SERVICE.main begin")
			end
			callback_context := application.service_callback_context (Current)
			status_handle := register_service_ctrl_handler_ex_a_external (string_to_pointer (name), application.service_c_dispatcher, callback_context)
			if status_handle /= default_pointer then
				set_status (service_start_pending)
				run_thread
				debug ("service")
					log ("NT_SERVICE.main end")
				end
				set_status (service_stopped)
			else
				debug ("service")
					log ("NT_SERVICE.main: register ctrl handler failed")
				end
			end
		ensure
			callback_context_set:
				callback_context /= default_pointer and then
				callback_context = application.service_callback_context (Current)
		rescue
			callback_context := default_pointer
			debug ("service")
				log ("NT_SERVICE.main: rescue")
				if exception_trace /= Void then
					log (exception_trace)
				end
			end
		end


feature {NT_SERVICE_THREAD} -- Service commands

	execute is
			-- Feature where service will spend most of its time.
			-- Service should call `process_requests' regularly so it react
			-- to SCM control requests in a timely manner.
		require
			thread_not_void: thread /= Void
			service_running: current_state = service_running
		do
			debug ("service")
				log ("NT_SERVICE.execute begin")
			end
			process_requests (True)
			debug ("service")
				log ("NT_SERVICE.execute end")
			end
		end

	interrogate is
		do
			debug ("service")
				log ("NT_SERVICE.interrogate")
			end
			report_status
		end

	shutdown is
		require
			thread_not_void: thread /= Void
		do
			debug ("service")
				log ("NT_SERVICE.shutdown")
			end
			set_status (service_stop_pending)
			thread.terminate
		ensure
			stop_pending: current_state /= service_stop_pending
			thread_terminated: thread.terminated
		end

	start is
			-- Initialization code for the service.
			-- Start should finish before `default_wait_hint'
			-- milliseconds have passed, or else it should regularly
			-- update its status.
			-- Should set `current_state' to `service_running' if
			-- successful, or `service_stop_pending' otherwise.
		require
			thread_not_void: thread /= Void
			start_pending: current_state = service_start_pending
		do
			set_status (service_running)
		ensure
			running_or_stopping: current_state = service_running or else current_state = service_stop_pending
		end

	stop is
			-- Shutdown code for the service.
		require
			thread_not_void: thread /= Void
			not_stopped: current_state /= service_stopped
			stop_not_pending: current_state /= service_stop_pending
		do
			debug ("service")
				log ("NT_SERVICE.stop: begin")
			end
			set_status (service_stop_pending)
			thread.terminate
			debug ("service")
				log ("NT_SERVICE.stop: end")
			end
		ensure
			stop_pending: current_state = service_stop_pending
			thread_terminated: thread.terminated
		end


feature {NONE} -- Service thread

	process_requests (wait_for_message: BOOLEAN) is
			-- Read message queue to be able to respond to commands send
			-- by the service controller.
		require
			thread_not_void: thread /= Void
		do
			thread.process_requests (wait_for_message)
		end

	run_thread is
			-- Routine may never raise an exception.
		require
			thread_void: thread = Void
		local
			retried: BOOLEAN
		do
			if not retried then
				debug ("service")
					log ("NT_SERVICE.run_thread: create thread")
				end
				create thread.make (Current)
				thread.launch
				debug ("service")
					log ("NT_SERVICE.run_thread: before join")
				end
				thread.join
				debug ("service")
					log ("NT_SERVICE.run_thread: after join")
				end
			end
			thread := Void
		ensure
			thread_void: thread = Void
		rescue
			retried := True
				debug ("service")
					log ("NT_SERVICE.run_thread: rescue")
				end
			retry
		end


feature {NT_SERVICE_THREAD} -- Service thread

	thread: NT_SERVICE_THREAD
			-- Thread that does the actual work done by this service.


feature {NT_SERVICE_APPLICATION} -- Controller callbacks

	controller (a_dwcontrol: INTEGER; a_dweventtype: INTEGER; a_lpeventdata: POINTER): INTEGER is
			-- Controller handler function, called by NT in thread that
			-- has called `StartServiceCtrlDispatcherEx', usually the main
			-- process thread. Should communicate `a_dwcontrol' to the
			-- service thread.
		local
			success: BOOLEAN
			error: INTEGER
		do
			-- We always return NO_ERROR. Should return
			-- ERROR_CALL_NOT_IMPLEMENTED if we don't handle a particular
			-- `a_dwcontrol'.
			if thread /= Void then
				success := cwin_post_thread_message (thread.thread_id, cm_service_control_code, a_dwcontrol, 0)
				if not success then
					error := last_error_code
					debug ("service")
						log ("NT_SERVICE.controller: sending msg to thread failed: " + error.out)
					end
				end
			end
		rescue
			debug ("service")
				log ("NT_SERVICE.controller: sending msg to thread failed: " + error.out)
			end
		end


invariant

	status_not_void: status /= Void

	application_not_void: application /= Void

	callback_context_correct:
		callback_context /= default_pointer implies
			callback_context = application.service_callback_context (Current)

end
