indexing

	description: "Root class for an NT executable that contains one or more services."

	author: "Berend de Boer"
	copyright:   "Copyright (c) 2003, Berend de Boer"
	license:     "Eiffel Forum Freeware License v2 (see forum.txt)"


deferred class

	NT_SERVICE_APPLICATION


inherit

	ANY

	NT_SERVICE_DEBUGGING
		export
			{NONE} all
		end

	WINSVC_FUNCTIONS_EXTERNAL
		export
			{NONE} all
		end

	MEMORY
		export
			{NONE} all
		end

	EXCEPTIONS
		export
			{NONE} all
		end

	LPSERVICE_MAIN_FUNCTIONA_CALLBACK
		rename
			on_callback as service_main
		end

	LPHANDLER_FUNCTION_EX_CALLBACK
		rename
			on_callback as controller
		end

	EWG_IMPORTED_EXTERNAL_ROUTINES
		export
			{NONE} all
		end

	WEL_WINDOWS_ERROR_MESSAGES
		export
			{NONE} all
		end


feature {NONE} -- Initialization

	make is
			-- Initialize NT service application and start services.
		local
			service_table: SERVICE_TABLE_ENTRYA_ARRAY
			r: INTEGER
			i: INTEGER
		do
			create_services
			create service_main_callback.make (Current)
			create controller_callback.make (Current)
			debug ("service")
				log ("NT_SERVICE_APPLICATION.make: creating service table")
			end
			create service_table.make_new_unshared (services.count + 1)
			from
				i := services.lower
			until
				i > services.upper
			loop
				service_table.item (i - services.lower).set_service_name (services.item (i).name)
				service_table.item (i - services.lower).set_lpserviceproc (service_main_callback.c_dispatcher)
				i := i + 1
			end
			debug ("service")
				log ("NT_SERVICE_APPLICATION.make: before ctrl dispatcher")
			end
			r := start_service_ctrl_dispatcher_a_external (service_table.array_address)
			cwin_sleep (2000)
			debug ("service")
				log ("NT_SERVICE_APPLICATION.make: after ctrl dispatcher")
			end
		rescue
			debug ("service")
				log ("NT_SERVICE_APPLICATION.make: rescue")
				if exception_trace /= Void then
					log (exception_trace)
				end
			end
		end


feature -- Access

	service_c_dispatcher: POINTER is
			-- Routine that service can use to pass as callback pointer
			-- to `test'.
		do
			Result := controller_callback.c_dispatcher
		ensure
			dispatcher_not_nil: Result /= default_pointer
		end

	service_callback_context (a_service: NT_SERVICE): POINTER is
			-- Context that can be passed to `RegisterServiceCtrlHandlerEx'.
		require
			service_added: services.has (a_service)
		do
			Result := EXTERNAL_MEMORY_.integer_to_pointer_external (array_index_of (services, a_service))
		ensure
			context_not_nil: Result /= default_pointer
		end

	services: ARRAY [NT_SERVICE]
			-- List of services to start.
			-- Only one supported for now.


feature {NONE} -- Callback

	service_main_callback: LPSERVICE_MAIN_FUNCTIONA_DISPATCHER
			-- Callback in separate thread to start a service.

	dispatch_to_service (arguments: ARRAY [STRING]) is
		do
			debug ("service")
				log ("NT_SERVICE_APPLICATION.dispatch_to_service: begin")
			end
			services.item (services.lower).main (arguments)
			debug ("service")
				log ("NT_SERVICE_APPLICATION.dispatch_to_service: end")
			end
		end

	service_main (dwnumservicesargs: INTEGER; lpserviceargvectors: POINTER) is
			-- Dispatch here to proper service based on first string in
			-- `lpserviceargvectors'.
			-- This routine is called in a separate thread, created by the SCM.
		local
			retried: BOOLEAN
		do
			if not retried then
				debug ("service")
					log ("NT_SERVICE_APPLICATION.service_main: begin")
				end
				dispatch_to_service (Void)
				debug ("service")
					log ("NT_SERVICE_APPLICATION.service_main: end")
				end
			end
		rescue
			if not retried then
				debug ("service")
					log ("NT_SERVICE_APPLICATION.service_main: rescue")
					if exception_trace /= Void then
						log (exception_trace)
					end
				end
				-- Make sure we never pass exceptions
				retried := True
				retry
			end
		end


feature {NONE} -- Callback

	controller_callback: LPHANDLER_FUNCTION_EX_DISPATCHER
			-- Callback function for individual services. Callback occurs
			-- to `controller'.

	controller (a_dwcontrol: INTEGER; a_dweventtype: INTEGER; a_lpeventdata: POINTER; a_lpcontext: POINTER): INTEGER is
			-- Controller handler function, called by SCM in thread that
			-- has called `StartServiceCtrlDispatcherEx', usually the
			-- main process thread. Should communicate `a_dwcontrol' to
			-- the service thread of the service in `a_lpcontext'.
			-- Due to limitations in EWG there is only a single callback
			-- function per application. The SCM needs a separate
			-- function for every service though. This function routes
			-- the call to the appropriate service.
		local
			index: INTEGER
			service: NT_SERVICE
			retried: BOOLEAN
		do
			if not retried then
				index := EXTERNAL_MEMORY_.pointer_to_integer_external (a_lpcontext)
				if index >= services.lower and index <= services.upper then
					service := services.item (index)
						check
							context_ok: service.callback_context = a_lpcontext
						end
					Result := service.controller (a_dwcontrol, a_dweventtype, a_lpeventdata)
				else
					debug ("service")
						log ("NT_SERVICE_APPLICATION.controller: given context did not resolve to a service.")
					end
					Result := Error_call_not_implemented_value
				end
			end
		rescue
			if not retried then
				debug ("service")
					log ("NT_SERVICE_APPLICATION.controller: call to service thread failed.")
				end
				retried := False
				retry
			end
		end


feature {NONE} -- Services

	create_services is
			-- Create and fill `services'.
		deferred
		ensure
			services_not_void: services /= Void
			one_service_supported: services.count = 1
			context_callback_available: service_callback_context (services.item (services.lower)) /= default_pointer
		end


feature {NONE} -- Implementation

	array_index_of (a: ARRAY [NT_SERVICE]; v: NT_SERVICE): INTEGER is
			-- Index of `v' in `a', given that `v' exists in `a'.
		require
			a_not_void: a /= Void
			v_exists: a.has (v)
		do
			from
				Result := a.lower
			invariant
				Result >= a.lower and then Result <= a.upper
			variant
				a.lower + (a.upper - Result)
			until
				v = a.item (Result)
			loop
				Result := Result + 1
			end
		ensure
			found: a.item (Result) = v
		end


feature {NONE} -- Singleton

	frozen singleton: NT_SERVICE_APPLICATION is
			-- This class.
		once
			Result := Current
		ensure
			returns_current: Result = Current
		end


invariant

	singleton: Current = singleton

	service_main_callback_not_void: service_main_callback /= Void
	controller_callback_not_void: controller_callback /= Void

	services_not_void: services /= Void
	only_one_service: services.count <= 1

end
